/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.xdomea;

import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.springframework.stereotype.Component;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.eingang.xta.EingangRepresentationsAndAttachments;
import de.ozgcloud.eingang.xta.XtaIncomingFilesMapper;
import de.ozgcloud.eingang.xta.zip.ZipFileExtractor;
import de.ozgcloud.xta.client.model.XtaMessage;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Component
@Log4j2
@RequiredArgsConstructor
public class XdomeaMessageDataMapper {
	static final String XDOMEA_0201_XML_FILE_NAME_SUFFIX = "_Geschaeftsgang.Geschaeftsgang.0201.xml";
	static final String ATTACHMENT_GROUP_NAME = "Dokument";

	private final XdomeaXMLValueReader valueReader;
	private final XtaIncomingFilesMapper xtaIncomingFilesMapper;
	private final ZipFileExtractor zipFileExtractor;

	public EingangRepresentationsAndAttachments extractEingangRepresentationsAndAttachments(XtaMessage message) {
		var incomingZipFile = xtaIncomingFilesMapper.toIncomingFile(message.messageFile());
		try {
			return tryExtractEingangRepresentationsAndAttachments(incomingZipFile);
		} catch (RuntimeException e) {
			LOG.error("Error while extracting Xdomea ZIP file!", e);
			return EingangRepresentationsAndAttachments.builder()
					.primaryFormDataFileName("")
					.primaryFormDataPdfFileName("")
					.representations(List.of(incomingZipFile))
					.attachments(List.of())
					.build();
		}
	}

	EingangRepresentationsAndAttachments tryExtractEingangRepresentationsAndAttachments(IncomingFile zipFile) {
		return mapEingangRepresentationsAndAttachments(
				zipFileExtractor.extractIncomingFilesSafely(zipFile)
		);
	}

	EingangRepresentationsAndAttachments mapEingangRepresentationsAndAttachments(List<IncomingFile> incomingFileList) {
		return removeUuidPrefixFromEingangFiles(
				mapRawEingangRepresentationsAndAttachments(incomingFileList)
		);
	}

	private EingangRepresentationsAndAttachments removeUuidPrefixFromEingangFiles(EingangRepresentationsAndAttachments eingang) {
		return EingangRepresentationsAndAttachments.builder()
				.primaryFormDataFileName(removeUuidPrefix(eingang.primaryFormDataFileName()))
				.primaryFormDataPdfFileName(removeUuidPrefix(eingang.primaryFormDataPdfFileName()))
				.representations(removeUuidPrefixFromIncomingFiles(eingang.representations()))
				.attachments(removeUuidPrefixFromIncomingFileGroups(eingang.attachments()))
				.build();
	}

	private EingangRepresentationsAndAttachments mapRawEingangRepresentationsAndAttachments(List<IncomingFile> incomingFileList) {
		var xdomeaMessageData = mapIncomingFilesToXdomeaMessageData(incomingFileList);
		return EingangRepresentationsAndAttachments.builder()
				.primaryFormDataFileName(findRepresentationFileNameBySuffixFromMetadata(xdomeaMessageData, ".xml"))
				.primaryFormDataPdfFileName(findRepresentationFileNameBySuffixFromMetadata(xdomeaMessageData, ".pdf"))
				.representations(getPairRepresentations(xdomeaMessageData))
				.attachments(xdomeaMessageData.attachments())
				.build();
	}

	private String findRepresentationFileNameBySuffixFromMetadata(XdomeaMessageData xdomeaMessageData, String suffix) {
		return findRepresentationFileNameBySuffix(
				xdomeaMessageData.representations().stream()
						.map(IncomingFile::getName)
						.toList(),
				suffix
		);
	}

	private List<IncomingFileGroup> removeUuidPrefixFromIncomingFileGroups(List<IncomingFileGroup> incomingFileGroups) {
		return incomingFileGroups.stream()
				.map(group -> group.toBuilder()
						.clearFiles()
						.files(removeUuidPrefixFromIncomingFiles(group.getFiles()))
						.build())
				.toList();
	}

	private List<IncomingFile> removeUuidPrefixFromIncomingFiles(List<IncomingFile> incomingFiles) {
		return incomingFiles.stream()
				.map(this::removeUuidPrefixFromIncomingFile)
				.toList();
	}

	private IncomingFile removeUuidPrefixFromIncomingFile(IncomingFile incomingFile) {
		return incomingFile.toBuilder()
				.name(removeUuidPrefix(incomingFile.getName()))
				.build();
	}

	private String removeUuidPrefix(String name) {
		return name.replaceFirst("^[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}_", "");
	}

	private List<IncomingFile> getPairRepresentations(XdomeaMessageData xdomeaMessageData) {
		return Stream.concat(
				Stream.of(xdomeaMessageData.metadataFile()),
				xdomeaMessageData.representations().stream()
		).toList();
	}

	XdomeaMessageData mapIncomingFilesToXdomeaMessageData(List<IncomingFile> incomingFileList) {
		var xdomeaXMLFile = findXdomeaXMLFile(incomingFileList);
		var representationFileNames = valueReader.readRepresentationFileNames(xdomeaXMLFile);
		var attachmentFileNames = getFileNamesExcluding(
				incomingFileList,
				Stream.concat(Stream.of(xdomeaXMLFile.getName()), representationFileNames.stream())
		);

		var fileNameToFileMap = createFileNameToIncomingFileMap(incomingFileList);
		return XdomeaMessageData.builder()
				.metadataFile(xdomeaXMLFile)
				.representations(getRepresentations(representationFileNames, fileNameToFileMap))
				.attachments(getAttachments(attachmentFileNames, fileNameToFileMap))
				.build();
	}

	private Map<String, IncomingFile> createFileNameToIncomingFileMap(List<IncomingFile> incomingFileList) {
		return incomingFileList.stream()
				.collect(Collectors.toMap(IncomingFile::getName, file -> file));
	}

	private Stream<String> getFileNamesExcluding(List<IncomingFile> incomingFileList, Stream<String> excludedFileNames) {
		return dropNames(
				incomingFileList.stream().map(IncomingFile::getName),
				excludedFileNames.collect(Collectors.toSet())
		);
	}

	private List<IncomingFile> getRepresentations(List<String> representationFileNames, Map<String, IncomingFile> fileNameToFileMap) {
		return getFilesByName(representationFileNames.stream(), fileNameToFileMap);
	}

	private List<IncomingFileGroup> getAttachments(Stream<String> attachmentFileNames, Map<String, IncomingFile> fileNameToFileMap) {
		return Stream.of(IncomingFileGroup.builder()
						.name(ATTACHMENT_GROUP_NAME)
						.files(
								getFilesByName(
										attachmentFileNames,
										fileNameToFileMap
								)
						)
						.build())
				// Do not include empty attachment group
				.filter(group -> !group.getFiles().isEmpty())
				.toList();
	}

	private List<IncomingFile> getFilesByName(Stream<String> names, Map<String, IncomingFile> fileNameToFileMap) {
		return names.map(fileNameToFileMap::get).toList();
	}

	private Stream<String> dropNames(Stream<String> names, Set<String> namesToDrop) {
		return names.filter(name -> !namesToDrop.contains(name));
	}

	String findRepresentationFileNameBySuffix(List<String> representationFileNames, String suffix) {
		var fileNames = representationFileNames.stream()
				.filter(name -> name.endsWith(suffix))
				.toList();
		if (fileNames.isEmpty()) {
			LOG.warn("There is no representation with suffix {}!.", suffix);
			return "";
		} else if (fileNames.size() > 1) {
			LOG.warn("There is more than one representation with suffix {}. Choosing the first of {}.", suffix, fileNames.size());
		}
		return fileNames.getFirst();
	}

	IncomingFile findXdomeaXMLFile(List<IncomingFile> incomingFileList) {
		return incomingFileList.stream()
				.filter(file -> file.getName().endsWith(XDOMEA_0201_XML_FILE_NAME_SUFFIX))
				.findFirst()
				.orElseThrow(() -> new TechnicalException("Primary xdomea xml representation not found!"));
	}

}
