/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.xta;

import java.util.List;
import java.util.Optional;
import java.util.function.Consumer;

import org.apache.commons.lang3.StringUtils;
import org.springframework.stereotype.Service;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.vorgang.VorgangNummerSupplier;
import de.ozgcloud.eingang.fim.FimMessageDataMapper;
import de.ozgcloud.eingang.semantik.SemantikAdapter;
import de.ozgcloud.eingang.xdomea.XdomeaMessageDataMapper;
import de.ozgcloud.xta.client.XtaClient;
import de.ozgcloud.xta.client.exception.XtaClientException;
import de.ozgcloud.xta.client.model.XtaMessage;
import de.ozgcloud.xta.client.model.XtaMessageStatus;
import de.ozgcloud.xta.client.model.XtaTransportReport;
import lombok.NonNull;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Service
@Log4j2
@RequiredArgsConstructor
class XtaService {

	static final String XDOMEA_0201_MESSAGE_TYPE = "Geschaeftsgang.Geschaeftsgang.0201";

	static final String FIM_MESSAGE_TYPE_PREFIX = "fim.";

	final Consumer<XtaMessage> messageProcessor = this::processMessage;

	private final XtaClient xtaClient;
	private final SemantikAdapter semantikAdapter;

	private final XtaMessageMapper mapper;
	private final VorgangNummerSupplier vorgangNummerSupplier;
	private final XdomeaMessageDataMapper xdomeaMessageDataMapper;
	private final FimMessageDataMapper fimMessageDataMapper;

	public void fetchMessages() throws XtaClientException {

		var transportReports = xtaClient.fetchMessages(messageProcessor);
		logTransportReports(transportReports);
	}

	void processMessage(XtaMessage xtaMessage) {
		Optional<FormData> formData = getFormDataIfNoRuntimeException(xtaMessage);
		formData.ifPresent(this::processSemantik);
	}

	Optional<FormData> getFormDataIfNoRuntimeException(@NonNull XtaMessage xtaMessage) {
		try {
			return Optional.of(getFormData(xtaMessage));
		} catch (RuntimeException exception) {
			LOG.error("Failed to process xta message with id: {}", xtaMessage.metaData().messageId(), exception);
			return Optional.empty();
		}
	}

	void processSemantik(@NonNull FormData formData) {
		try {
			LOG.info("Semantic processing XTA-Message '{}'.", formData.getHeader().getRequestId());
			semantikAdapter.processFormData(formData);
		} catch (RuntimeException e) {
			LOG.error("Error on semantic processing XTA-Message. Continue with next message.", e);
		}
	}

	FormData getFormData(XtaMessage xtaMessage) {
		return mapper.toFormData(
				getRepresentationsAndAttachments(xtaMessage),
				xtaMessage.metaData(),
				vorgangNummerSupplier
		);
	}

	EingangRepresentationsAndAttachments getRepresentationsAndAttachments(XtaMessage message) {
		var metaData = message.metaData();
		if (isXDomeaMessageType(metaData.messageTypeCode())) {
			return xdomeaMessageDataMapper.extractEingangRepresentationsAndAttachments(message);
		}
		if (isFimMessageType(metaData.messageTypeCode())) {
			return fimMessageDataMapper.mapEingangRepresentationsAndAttachments(message);
		}

		throw new TechnicalException("Unexpected XTA message type: %s".formatted(metaData.messageTypeCode()));
	}

	private boolean isXDomeaMessageType(String messageType) {
		return StringUtils.equals(messageType, XDOMEA_0201_MESSAGE_TYPE);
	}

	private boolean isFimMessageType(String messageType) {
		return StringUtils.startsWith(messageType, FIM_MESSAGE_TYPE_PREFIX);
	}

	void logTransportReports(List<XtaTransportReport> transportReports) {
		for (var transportReport : transportReports) {
			if (transportReport.status() == XtaMessageStatus.GREEN) {
				LOG.info("Fetched XTA message: {}", transportReport.metaData().messageId());
			} else {
				LOG.warn(
						"Failed to fetch XTA message with XTA transport report status {}: {}",
						transportReport.status(),
						transportReport.metaData().messageId()
				);
			}
		}
	}
}
