package de.ozgcloud.eingang.fim;

import static de.ozgcloud.eingang.fim.FimMessageDataMapper.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.eingang.xta.EingangRepresentationsAndAttachments;
import de.ozgcloud.eingang.xta.XtaFileTestFactory;
import de.ozgcloud.eingang.xta.XtaIncomingFilesMapper;
import de.ozgcloud.eingang.xta.XtaMessageTestFactory;
import de.ozgcloud.xta.client.model.XtaFile;
import de.ozgcloud.xta.client.model.XtaMessage;

class FimMessageDataMapperTest {

	@Spy
	@InjectMocks
	private FimMessageDataMapper mapper;

	@Mock
	private XtaIncomingFilesMapper xtaIncomingFilesMapper;

	private final String primaryFormDataFileName = "test-fim.xml";
	private final String fimAttachmentFileName = "upload.pdf";
	private final String fimAttachment2FileName = "upload.png";

	private final XtaFile primaryFormDataFile = XtaFileTestFactory.createBuilder()
			.name(primaryFormDataFileName)
			.build();

	private final XtaFile fimAttachment1 = XtaFileTestFactory.createBuilder()
			.name(fimAttachmentFileName)
			.build();

	private final XtaFile fimAttachment2 = XtaFileTestFactory.createBuilder()
			.name(fimAttachment2FileName)
			.build();

	private final XtaMessage message = XtaMessageTestFactory.createBuilder()
			.messageFile(primaryFormDataFile)
			.attachmentFiles(List.of(fimAttachment1, fimAttachment2))
			.build();

	@DisplayName("map EingangRepresentationsAndAttachments")
	@Nested
	class TestMapEingangRepresentationsAndAttachments {

		@Mock
		private IncomingFile representationIncomingFile;

		@Mock
		private IncomingFileGroup attachmentIncomingFileGroup;

		@BeforeEach
		void mock() {
			doReturn(List.of(representationIncomingFile)).when(mapper).getRepresentations(any());
			doReturn(List.of(attachmentIncomingFileGroup)).when(mapper).getAttachmentGroups(any());
		}

		@DisplayName("should map primaryFormDataFileName")
		@Test
		void shouldMapPrimaryFormDataFileName() {
			var result = mapEingangRepresentationsAndAttachments();

			assertThat(result.primaryFormDataFileName()).isEqualTo(primaryFormDataFileName);
		}

		@DisplayName("should map primaryFormDataPdfFileName")
		@Test
		void shouldMapPrimaryFormDataPdfFileName() {
			var result = mapEingangRepresentationsAndAttachments();

			assertThat(result.primaryFormDataPdfFileName()).isEmpty();
		}

		@DisplayName("should call getRepresentations")
		@Test
		void shouldCallGetRepresentations() {
			mapEingangRepresentationsAndAttachments();

			verify(mapper).getRepresentations(message);
		}

		@DisplayName("should map representations")
		@Test
		void shouldMapRepresentations() {
			var result = mapEingangRepresentationsAndAttachments();

			assertThat(result.representations()).containsExactly(representationIncomingFile);
		}

		@DisplayName("should call getAttachmentGroups")
		@Test
		void shouldCallGetAttachmentGroups() {
			mapEingangRepresentationsAndAttachments();

			verify(mapper).getAttachmentGroups(message);
		}

		@DisplayName("should map attachments")
		@Test
		void shouldMapAttachments() {
			var result = mapEingangRepresentationsAndAttachments();

			assertThat(result.attachments()).containsExactly(attachmentIncomingFileGroup);
		}

		private EingangRepresentationsAndAttachments mapEingangRepresentationsAndAttachments() {
			return mapper.mapEingangRepresentationsAndAttachments(message);
		}
	}

	@DisplayName("get attachment groups")
	@Nested
	class TestGetAttachmentGroups {
		@Mock
		private IncomingFile attachmentIncomingFile1;

		@Mock
		private IncomingFile attachmentIncomingFile2;

		@BeforeEach
		void mock() {
			when(xtaIncomingFilesMapper.toIncomingFile(any()))
					.thenReturn(attachmentIncomingFile1)
					.thenReturn(attachmentIncomingFile2);

		}

		@DisplayName("should call toIncomingFile")
		@Test
		void shouldCallToIncomingFile() {
			mapper.getAttachmentGroups(message);

			verify(xtaIncomingFilesMapper).toIncomingFile(fimAttachment1);
			verify(xtaIncomingFilesMapper).toIncomingFile(fimAttachment2);
		}

		@DisplayName("should return group with name")
		@Test
		void shouldReturnGroupWithName() {
			var result = mapper.getAttachmentGroups(message);

			assertThat(result)
					.extracting(IncomingFileGroup::getName)
					.containsExactly(FIM_ATTACHMENT_GROUP_NAME);
		}

		@DisplayName("should return group with files")
		@Test
		void shouldReturnGroupWithFiles() {
			var result = mapper.getAttachmentGroups(message);

			assertThat(result)
					.flatExtracting(IncomingFileGroup::getFiles)
					.containsExactly(attachmentIncomingFile1, attachmentIncomingFile2);
		}
	}

	@DisplayName("get representations")
	@Nested
	class TestGetRepresentations {
		@Mock
		private IncomingFile representationIncomingFile;

		@BeforeEach
		void mock() {
			when(xtaIncomingFilesMapper.toIncomingFile(any())).thenReturn(representationIncomingFile);
		}

		@DisplayName("should call toIncomingFile")
		@Test
		void shouldCallToIncomingFile() {
			mapper.getRepresentations(message);

			verify(xtaIncomingFilesMapper).toIncomingFile(primaryFormDataFile);
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var result = mapper.getRepresentations(message);

			assertThat(result).containsExactly(representationIncomingFile);
		}
	}

}