/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.xdomea;

import static de.ozgcloud.eingang.xdomea.XdomeaMessageDataMapper.*;
import static de.ozgcloud.eingang.xdomea.XdomeaMessageDataTestFactory.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.Collection;
import java.util.List;
import java.util.stream.Stream;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.IncomingFile;
import de.ozgcloud.eingang.common.formdata.IncomingFileGroup;
import de.ozgcloud.eingang.common.formdata.IncomingFileTestFactory;
import de.ozgcloud.eingang.xta.EingangRepresentationsAndAttachments;
import de.ozgcloud.eingang.xta.XtaIncomingFilesMapper;
import de.ozgcloud.eingang.xta.XtaMessageTestFactory;
import de.ozgcloud.eingang.xta.zip.ZipFileExtractor;
import de.ozgcloud.xta.client.model.XtaMessage;

class XdomeaMessageDataMapperTest {

	@Spy
	@InjectMocks
	private XdomeaMessageDataMapper mapper;

	@Mock
	private XdomeaXMLValueReader valueReader;

	@Mock
	private ZipFileExtractor zipFileExtractor;

	@Mock
	private XtaIncomingFilesMapper xtaIncomingFilesMapper;

	@DisplayName("extract EingangRepresentationsAndAttachments")
	@Nested
	class TestExtractEingangRepresentationsAndAttachments {

		private final XtaMessage message = XtaMessageTestFactory.create();

		@Mock
		private IncomingFile zipFile;

		@Mock
		private EingangRepresentationsAndAttachments eingang;

		@BeforeEach
		void mock() {
			when(xtaIncomingFilesMapper.toIncomingFile(any())).thenReturn(zipFile);
		}

		@DisplayName("without exception")
		@Nested
		class TestWithoutException {
			@BeforeEach
			void mock() {
				doReturn(eingang).when(mapper).tryExtractEingangRepresentationsAndAttachments(any());
			}

			@DisplayName("should call toIncomingFile")
			@Test
			void shouldCallToIncomingFile() {
				extractEingangRepresentationsAndAttachments();

				verify(xtaIncomingFilesMapper).toIncomingFile(message.messageFile());
			}

			@DisplayName("should call tryExtractEingangRepresentationsAndAttachments")
			@Test
			void shouldCallTryExtractEingangRepresentationsAndAttachments() {
				extractEingangRepresentationsAndAttachments();

				verify(mapper).tryExtractEingangRepresentationsAndAttachments(zipFile);
			}

			@DisplayName("should return")
			@Test
			void shouldReturn() {
				var result = extractEingangRepresentationsAndAttachments();

				assertThat(result).isEqualTo(eingang);
			}
		}

		@DisplayName("with exception")
		@Nested
		class TestWithException {
			@Mock
			private RuntimeException exception;

			@BeforeEach
			void mock() {
				doThrow(exception).when(mapper).tryExtractEingangRepresentationsAndAttachments(any());
			}

			@DisplayName("should map primaryFormDataFileName to empty")
			@Test
			void shouldMapPrimaryFormDataFileNameToEmpty() {
				var result = extractEingangRepresentationsAndAttachments();

				assertThat(result.primaryFormDataFileName()).isEmpty();
			}

			@DisplayName("should map primaryFormDataPdfFileName to empty")
			@Test
			void shouldMapPrimaryFormDataPdfFileNameToEmpty() {
				var result = extractEingangRepresentationsAndAttachments();

				assertThat(result.primaryFormDataPdfFileName()).isEmpty();
			}

			@DisplayName("should map representations")
			@Test
			void shouldMapRepresentations() {
				var result = extractEingangRepresentationsAndAttachments();

				assertThat(result.representations()).containsExactly(zipFile);
			}

			@DisplayName("should map attachments to empty")
			@Test
			void shouldMapAttachmentsToEmpty() {
				var result = extractEingangRepresentationsAndAttachments();

				assertThat(result.attachments()).isEmpty();
			}
		}

		private EingangRepresentationsAndAttachments extractEingangRepresentationsAndAttachments() {
			return mapper.extractEingangRepresentationsAndAttachments(message);
		}
	}

	@DisplayName("try extract EingangRepresentationsAndAttachments")
	@Nested
	class TestTryExtractEingangRepresentationsAndAttachments {

		@Mock
		private IncomingFile zipFile;

		@Mock
		private IncomingFile xdomeaXmlFile;

		@Mock
		private EingangRepresentationsAndAttachments eingang;

		@BeforeEach
		void mock() {
			when(zipFileExtractor.extractIncomingFilesSafely(any())).thenReturn(List.of(xdomeaXmlFile));
			doReturn(eingang).when(mapper).mapEingangRepresentationsAndAttachments(any());
		}

		@DisplayName("should call extractIncomingFilesSafely")
		@Test
		void shouldCallExtractIncomingFilesSafely() {
			tryExtractEingangRepresentationsAndAttachments();

			verify(zipFileExtractor).extractIncomingFilesSafely(zipFile);
		}

		@DisplayName("should call mapEingangRepresentationsAndAttachments")
		@Test
		void shouldCallMapEingangRepresentationsAndAttachments() {
			tryExtractEingangRepresentationsAndAttachments();

			verify(mapper).mapEingangRepresentationsAndAttachments(List.of(xdomeaXmlFile));
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var result = tryExtractEingangRepresentationsAndAttachments();

			assertThat(result).isEqualTo(eingang);
		}

		private EingangRepresentationsAndAttachments tryExtractEingangRepresentationsAndAttachments() {
			return mapper.tryExtractEingangRepresentationsAndAttachments(zipFile);
		}
	}

	@DisplayName("map EingangRepresentationsAndAttachments")
	@Nested
	class TestMapEingangRepresentationsAndAttachments {

		@Mock
		private List<IncomingFile> incomingFileList;

		@DisplayName("should map representations")
		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldMapRepresentations(boolean useUuidPrefix) {
			mockXdomeaMessageData(useUuidPrefix);

			var pair = mapper.mapEingangRepresentationsAndAttachments(incomingFileList);

			var representationFilenames = pair.representations().stream().map(IncomingFile::getName).toList();
			assertThat(representationFilenames).containsExactly(XDOMEA_XML_NAME, REPR_XML_NAME, REPR_PDF_NAME);
		}

		@DisplayName("should map attachments")
		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldMapAttachments(boolean useUuidPrefix) {
			mockXdomeaMessageData(useUuidPrefix);

			var pair = mapper.mapEingangRepresentationsAndAttachments(incomingFileList);

			var attachmentFilenames = pair.attachments().stream()
					.map(IncomingFileGroup::getFiles)
					.flatMap(Collection::stream)
					.map(IncomingFile::getName)
					.toList();
			assertThat(attachmentFilenames).containsExactly(ATTATCHMENT_XML_NAME, ATTATCHMENT_PNG_NAME, ATTATCHMENT_PDF_NAME);
		}

		@DisplayName("should map primaryFormDataFileName")
		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldMapPrimaryFormDataFileName(boolean useUuidPrefix) {
			mockXdomeaMessageData(useUuidPrefix);

			var pair = mapper.mapEingangRepresentationsAndAttachments(incomingFileList);

			assertThat(pair.primaryFormDataFileName()).isEqualTo(REPR_XML_NAME);
		}

		@DisplayName("should map primaryFormDataPdfFileName")
		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldMapPrimaryFormPdfDataFileName(boolean useUuidPrefix) {
			mockXdomeaMessageData(useUuidPrefix);

			var pair = mapper.mapEingangRepresentationsAndAttachments(incomingFileList);

			assertThat(pair.primaryFormDataPdfFileName()).isEqualTo(REPR_PDF_NAME);
		}

		private void mockXdomeaMessageData(boolean useUuidPrefix) {
			doReturn(XdomeaMessageDataTestFactory.create(useUuidPrefix)).when(mapper).mapIncomingFilesToXdomeaMessageData(incomingFileList);
		}

	}

	@DisplayName("find Xdomea XML file")
	@Nested
	class TestFindXdomeaXmlFile {

		private static final String FILE_NAME_WITHOUT_SUFFIX = "some-file-name.xml";
		private static final String FILE_NAME_WITH_SUFFIX = "some-file-name" + XDOMEA_0201_XML_FILE_NAME_SUFFIX;

		@DisplayName("should throw if not found")
		@Test
		void shouldThrowIfNotFound() {
			var incomingFilesWithout = List.of(IncomingFileTestFactory.createBuilder().name(FILE_NAME_WITHOUT_SUFFIX).build());

			assertThatThrownBy(() -> mapper.findXdomeaXMLFile(incomingFilesWithout))
					.isInstanceOf(TechnicalException.class);
		}

		@DisplayName("should return if found")
		@Test
		void shouldReturnIfFound() {
			var targetIncomingFile = IncomingFileTestFactory.createBuilder().name(FILE_NAME_WITH_SUFFIX).build();
			var incomingFilesWith = List.of(
					IncomingFileTestFactory.createBuilder().name(FILE_NAME_WITHOUT_SUFFIX).build(),
					targetIncomingFile
			);

			var primaryRepresentation = mapper.findXdomeaXMLFile(incomingFilesWith);

			assertThat(primaryRepresentation).isEqualTo(targetIncomingFile);
		}
	}

	@DisplayName("find representation file name by suffix")
	@Nested
	class TestFindRepresentationFileNameBySuffix {
		private static final String PDF_FILE_NAME = "some-file-name.pdf";
		private static final String PDF_FILE_NAME2 = "some-file-name.xml.pdf";
		private static final String XML_FILE_NAME = "some-file-name.xml";
		private static final String XML_FILE_NAME2 = "some-file-name.pdf.xml";

		@DisplayName("should return empty if no xml suffix")
		@Test
		void shouldReturnEmptyIfNoXmlSuffix() {
			var namesWithSuffix = List.of(PDF_FILE_NAME, PDF_FILE_NAME2);

			var fileName = mapper.findRepresentationFileNameBySuffix(namesWithSuffix, ".xml");

			assertThat(fileName).isEmpty();
		}

		@DisplayName("should return first with xml suffix")
		@Test
		void shouldReturnFirstWithXmlSuffix() {
			var namesWithSuffix = List.of(PDF_FILE_NAME, XML_FILE_NAME, PDF_FILE_NAME2,
					XML_FILE_NAME2);

			var fileName = mapper.findRepresentationFileNameBySuffix(namesWithSuffix, ".xml");

			assertThat(fileName).isEqualTo(XML_FILE_NAME);
		}

		@DisplayName("should return empty if no pdf suffix")
		@Test
		void shouldReturnEmptyIfNoPdfSuffix() {
			var namesWithoutSuffix = List.of(XML_FILE_NAME, XML_FILE_NAME2);

			var fileName = mapper.findRepresentationFileNameBySuffix(namesWithoutSuffix, ".pdf");

			assertThat(fileName).isEmpty();
		}

		@DisplayName("should return first with pdf suffix")
		@Test
		void shouldReturnFirstWithPdfSuffix() {
			var namesWithSuffix = List.of(PDF_FILE_NAME, XML_FILE_NAME, PDF_FILE_NAME2, XML_FILE_NAME2);

			var fileName = mapper.findRepresentationFileNameBySuffix(namesWithSuffix, ".pdf");

			assertThat(fileName).isEqualTo(PDF_FILE_NAME);
		}
	}

	@DisplayName("classify attachments and representations")
	@Nested
	class TestClassifyAttachmentsAndRepresentations {

		@Mock
		private IncomingFile xdomeaXMLFile;

		private List<IncomingFile> incomingFileList;

		@BeforeEach
		void mock() {
			incomingFileList = Stream.of(
							XDOMEA_XML_NAME,
							REPR_XML_NAME,
							REPR_PDF_NAME,
							ATTATCHMENT_XML_NAME,
							ATTATCHMENT_PNG_NAME,
							ATTATCHMENT_PDF_NAME
					)
					.map(name -> IncomingFileTestFactory.createBuilder().name(name).build())
					.toList();
			doReturn(xdomeaXMLFile).when(mapper).findXdomeaXMLFile(incomingFileList);

			when(xdomeaXMLFile.getName()).thenReturn(XDOMEA_XML_NAME);

			var representationFileNames = List.of(REPR_PDF_NAME, REPR_XML_NAME);
			when(valueReader.readRepresentationFileNames(xdomeaXMLFile)).thenReturn(representationFileNames);
		}

		@DisplayName("should contain xdomea metadata file")
		@Test
		void shouldContainXdomeaMetadataFile() {
			var classification = doClassify();

			var primaryDocument = classification.metadataFile();
			assertThat(primaryDocument.getName()).isEqualTo(XDOMEA_XML_NAME);
		}

		@DisplayName("should contain representations")
		@Test
		void shouldContainRepresentations() {
			var classification = doClassify();

			var resultRepresentationFileNames = classification.representations().stream()
					.map(IncomingFile::getName)
					.toList();

			assertThat(resultRepresentationFileNames).isEqualTo(List.of(REPR_PDF_NAME, REPR_XML_NAME));
		}

		@DisplayName("should contain attachments")
		@Test
		void shouldContainAttachments() {
			var classification = doClassify();

			var resultAttachmentFileNames = classification.attachments()
					.getFirst()
					.getFiles().stream()
					.map(IncomingFile::getName)
					.toList();
			assertThat(resultAttachmentFileNames).isEqualTo(List.of(ATTATCHMENT_XML_NAME, ATTATCHMENT_PNG_NAME, ATTATCHMENT_PDF_NAME));
		}

		@DisplayName("should have attachment group name")
		@Test
		void shouldHaveAttachmentGroupName() {
			var classification = doClassify();

			var resultAttachmentGroupName = classification.attachments()
					.getFirst()
					.getName();
			assertThat(resultAttachmentGroupName).isEqualTo(ATTACHMENT_GROUP_NAME);
		}

		private XdomeaMessageData doClassify() {
			return mapper.mapIncomingFilesToXdomeaMessageData(incomingFileList);
		}
	}

}
