/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.xta;

import static de.ozgcloud.eingang.xta.XtaMessageMapper.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mapstruct.factory.Mappers;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.formdata.FormMetaData;
import de.ozgcloud.eingang.common.formdata.FormMetaDataTestFactory;
import de.ozgcloud.eingang.common.vorgang.VorgangNummerSupplier;
import de.ozgcloud.xta.client.model.XtaMessageMetaData;

class XtaMessageMapperTest {

	@Spy
	@InjectMocks
	private XtaMessageMapper mapper = Mappers.getMapper(XtaMessageMapper.class);

	@Mock
	private VorgangNummerSupplier vorgangNummerSupplier;

	@Nested
	class TestToFormData {

		private XtaMessageMetaData xtaMessageMetaData;
		private EingangRepresentationsAndAttachments eingangRepresentationsAndAttachments;

		@BeforeEach
		void mock() {
			xtaMessageMetaData = XtaMessageMetaDataTestFactory.create();
			eingangRepresentationsAndAttachments = EingangRepresentationsAndAttachmentsTestFactory.create();
			when(vorgangNummerSupplier.get(VORGANG_NUMMER_SUFFIX_LENGTH)).thenReturn(FormHeaderTestFactory.VORGANGNUMMER);
		}

		@Test
		void shouldMapRepresentations() {
			var formData = doMapping();

			assertThat(formData.getRepresentations()).isEqualTo(eingangRepresentationsAndAttachments.representations());
		}

		@Test
		void shouldSetNumberOfAttachments() {
			var formData = doMapping();

			assertThat(formData.getNumberOfAttachments()).isEqualTo(3);
		}

		@Test
		void shouldSetNumberOfRepresentations() {
			var formData = doMapping();

			assertThat(formData.getNumberOfRepresentations()).isEqualTo(2);
		}

		@Test
		void shouldMapAttachments() {
			var formData = doMapping();

			assertThat(formData.getAttachments()).isEqualTo(eingangRepresentationsAndAttachments.attachments());
		}

		@Test
		void shouldMapVorgangNummer() {
			var formData = doMapping();

			assertThat(formData.getHeader().getVorgangNummer()).isEqualTo(FormHeaderTestFactory.VORGANGNUMMER);
		}

		@Test
		void shouldSetChannelXTA() {
			var formData = doMapping();

			assertThat(formData.getHeader().getChannel()).isEqualTo(FormHeaderTestFactory.CHANNEL);
		}

		@Test
		void shouldSetDesinationId() {
			var formData = doMapping();

			assertThat(formData.getControl().getMetaData()).isPresent().get()
					.extracting(metaData -> metaData.getEntry(XtaFormMetaData.XTA_IDENTIFIER_ENTRY_NAME))
					.isEqualTo(FormMetaDataTestFactory.XTA_IDENTIFIER);
		}

		@DisplayName("should map XtaFormMetaData service")
		@Test
		void shouldMapXtaFormMetaDataService() {
			var formData = doMapping();

			assertThat(formData.getControl().getMetaData()).isPresent().get()
					.extracting(metaData -> metaData.getEntry(XtaFormMetaData.SERVICE_ENTRY_NAME))
					.isEqualTo(xtaMessageMetaData.service());
		}

		@DisplayName("should map XtaFormMetaData origin")
		@Test
		void shouldMapXtaFormMetaDataOrigin() {
			var formData = doMapping();

			assertThat(formData.getControl().getMetaData()).isPresent().get()
					.extracting(FormMetaData::getOrigin)
					.isEqualTo(xtaMessageMetaData.deliveryAttributesOrigin());
		}

		@DisplayName("should map XtaFormMetaData delivery")
		@Test
		void shouldMapXtaFormMetaDataDelivery() {
			var formData = doMapping();

			assertThat(formData.getControl().getMetaData()).isPresent().get()
					.extracting(FormMetaData::getDelivery)
					.isEqualTo(xtaMessageMetaData.deliveryAttributesDelivery());
		}

		@DisplayName("should map XtaFormMetaData messageId")
		@Test
		void shouldMapXtaFormMetaDataMessageId() {
			var formData = doMapping();

			assertThat(formData.getControl().getMetaData()).isPresent().get()
					.extracting(metaData -> metaData.getEntry(XtaFormMetaData.MESSAGE_ID_ENTRY_NAME))
					.isEqualTo(xtaMessageMetaData.messageId());
		}

		@DisplayName("should map XtaFormMetaData messageType")
		@Test
		void shouldMapXtaFormMetaDataMessageType() {
			var formData = doMapping();

			assertThat(formData.getControl().getMetaData()).isPresent().get()
					.extracting(metaData -> metaData.getEntry(XtaFormMetaData.MESSAGE_TYPE_ENTRY_NAME))
					.isEqualTo(xtaMessageMetaData.messageTypeCode());
		}

		@DisplayName("should map XtaFormMetaData xtaIdentifier")
		@Test
		void shouldMapXtaFormMetaDataXtaIdentifier() {
			var formData = doMapping();

			assertThat(formData.getControl().getMetaData()).isPresent().get()
					.extracting(metaData -> metaData.getEntry(XtaFormMetaData.XTA_IDENTIFIER_ENTRY_NAME))
					.isEqualTo(xtaMessageMetaData.readerIdentifier().value());
		}

		@DisplayName("should map formEngineName")
		@Test
		void shouldMapFormEngineName() {
			var formData = doMapping();

			assertThat(formData.getHeader().getFormEngineName()).isEqualTo(FormHeaderTestFactory.XDOMEA_FORM_ENGINE_NAME);
		}

		@DisplayName("should map formEngineName for fim")
		@Test
		void shouldMapFormEngineNameForFim() {
			var formData = mapper.toFormData(eingangRepresentationsAndAttachments, XtaMessageMetaDataTestFactory.createFim(), vorgangNummerSupplier);

			assertThat(formData.getHeader().getFormEngineName()).isEqualTo(FormHeaderTestFactory.FIM_FORM_ENGINE_NAME);
		}

		private FormData doMapping() {
			return mapper.toFormData(eingangRepresentationsAndAttachments, xtaMessageMetaData, vorgangNummerSupplier);
		}
	}

}
