/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.eingang.xta;

import static de.ozgcloud.eingang.xta.XtaService.*;
import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Optional;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;

import de.ozgcloud.eingang.common.errorhandling.TechnicalException;
import de.ozgcloud.eingang.common.formdata.FormData;
import de.ozgcloud.eingang.common.vorgang.VorgangNummerSupplier;
import de.ozgcloud.eingang.fim.FimMessageDataMapper;
import de.ozgcloud.eingang.semantik.SemantikAdapter;
import de.ozgcloud.eingang.xdomea.XdomeaMessageDataMapper;
import de.ozgcloud.xta.client.XtaClient;
import de.ozgcloud.xta.client.model.XtaMessage;
import de.ozgcloud.xta.client.model.XtaMessageMetaData;
import de.ozgcloud.xta.client.model.XtaTransportReport;
import lombok.SneakyThrows;

class XtaServiceTest {

	@Spy
	@InjectMocks
	private XtaService service;

	@Mock
	private XtaClient xtaClient;
	@Mock
	private XtaMessageMapper mapper;
	@Mock
	private VorgangNummerSupplier vorgangNummerSupplier;
	@Mock
	private SemantikAdapter semantikAdapter;

	@Mock
	private XdomeaMessageDataMapper xdomeaMessageDataMapper;

	@Mock
	private FimMessageDataMapper fimMessageDataMapper;

	@Mock
	private FormData formData;

	@Nested
	class TestFetchMessages {

		private final XtaTransportReport transportReport = XtaTransportReportTestFactory.create();

		@BeforeEach
		@SneakyThrows
		void setup() {
			doReturn(List.of(transportReport)).when(xtaClient).fetchMessages(any());
		}

		@Test
		@SneakyThrows
		void shouldUseMessageProcessor() {
			service.fetchMessages();

			verify(xtaClient).fetchMessages(service.messageProcessor);
		}

		@Test
		@SneakyThrows
		void shouldCallLogTransportReports() {
			service.fetchMessages();

			verify(service).logTransportReports(any());
		}
	}

	@DisplayName("process message")
	@Nested
	class TestProcessMessage {
		private final XtaMessage xtaMessage = XtaMessageTestFactory.create();

		@DisplayName("should call getFormDataIfNoRuntimeException")
		@Test
		void shouldCallGetFormDataIfNoRuntimeException() {
			service.processMessage(xtaMessage);

			verify(service).getFormDataIfNoRuntimeException(any());
		}

		@Test
		void shouldCallProcessSemantik() {
			doReturn(Optional.of(formData)).when(service).getFormDataIfNoRuntimeException(any());

			service.processMessage(xtaMessage);

			verify(service).processSemantik(formData);
		}

		@Test
		void shouldNotCallProcessSemantik() {
			doReturn(Optional.empty()).when(service).getFormDataIfNoRuntimeException(any());

			service.processMessage(xtaMessage);

			verify(service, never()).processSemantik(any());
		}
	}

	@DisplayName("get form data if no runtime exception")
	@Nested
	class TestGetFormDataIfNoRuntimeException {

		XtaMessage message;

		@Mock
		FormData formData;

		@BeforeEach
		void before() {
			message = XtaMessageTestFactory.create();
		}

		@DisplayName("should call get formdata")
		@Test
		void shouldCallGetFormdata() {
			service.getFormDataIfNoRuntimeException(message);

			verify(service).getFormData(message);
		}

		@DisplayName("with exception")
		@Nested
		class TestWithException {
			@DisplayName("should return empty")
			@Test
			void shouldReturnEmpty() {
				doThrow(new RuntimeException("test-error")).when(service).getFormData(any());

				var result = service.getFormDataIfNoRuntimeException(message);

				assertThat(result).isEmpty();
			}
		}

		@DisplayName("without exception")
		@Nested
		class TestWithoutException {
			@DisplayName("should return")
			@Test
			void shouldReturn() {
				doReturn(formData).when(service).getFormData(any());

				var result = service.getFormDataIfNoRuntimeException(message);

				assertThat(result).contains(formData);
			}
		}
	}

	@DisplayName("get form data")
	@Nested
	class TestGetFormData {

		private final XtaMessage message = XtaMessageTestFactory.create();
		private final XtaMessageMetaData messageMetaData = XtaMessageMetaDataTestFactory.create();
		private final FormData mappedFormData = FormDataTestFactory.create();
		private final EingangRepresentationsAndAttachments eingang = EingangRepresentationsAndAttachmentsTestFactory.create();

		@BeforeEach
		void mock() {
			doReturn(eingang).when(service).getRepresentationsAndAttachments(any());
			when(mapper.toFormData(any(), any(), any())).thenReturn(mappedFormData);
		}

		@DisplayName("should call getRepresentationsAndAttachments")
		@Test
		void shouldCallGetRepresentationsAndAttachments() {
			service.getFormData(message);

			verify(service).getRepresentationsAndAttachments(message);
		}

		@DisplayName("should call toFormData")
		@Test
		void shouldCallToFormData() {
			service.getFormData(message);

			verify(mapper).toFormData(eingang, messageMetaData, vorgangNummerSupplier);
		}

		@DisplayName("should return")
		@Test
		void shouldReturn() {
			var result = service.getFormData(message);

			assertThat(result).isEqualTo(mappedFormData);
		}
	}

	@DisplayName("get representations and attachments")
	@Nested
	class TestGetRepresentationsAndAttachments {

		@Mock
		private EingangRepresentationsAndAttachments eingang;

		@DisplayName("with xdomea type")
		@Nested
		class TestWithXdomeaType {

			private final XtaMessage xdomeaMessage = XtaMessageTestFactory.createBuilder()
					.metaData(XtaMessageMetaDataTestFactory.createBuilder()
							.messageTypeCode(XtaService.XDOMEA_0201_MESSAGE_TYPE)
							.build())
					.build();

			@BeforeEach
			void mock() {
				when(xdomeaMessageDataMapper.extractEingangRepresentationsAndAttachments(any())).thenReturn(eingang);
			}

			@DisplayName("should call extractEingangRepresentationsAndAttachments")
			@Test
			void shouldCallExtractEingangRepresentationsAndAttachments() {
				service.getRepresentationsAndAttachments(xdomeaMessage);

				verify(xdomeaMessageDataMapper).extractEingangRepresentationsAndAttachments(xdomeaMessage);
			}

			@DisplayName("should return")
			@Test
			void shouldReturn() {
				var result = service.getRepresentationsAndAttachments(xdomeaMessage);

				assertThat(result).isEqualTo(eingang);
			}
		}

		@DisplayName("with fim type")
		@Nested
		class TestWithFimType {
			private final XtaMessage fimMessage = XtaMessageTestFactory.createBuilder()
					.metaData(XtaMessageMetaDataTestFactory.createBuilder()
							.messageTypeCode(FIM_MESSAGE_TYPE_PREFIX + "836487")
							.build())
					.build();

			@BeforeEach
			void mock() {
				when(fimMessageDataMapper.mapEingangRepresentationsAndAttachments(any())).thenReturn(eingang);
			}

			@DisplayName("should call mapIncomingFilesToRepresentations")
			@Test
			void shouldCallMapIncomingFilesToRepresentations() {
				service.getRepresentationsAndAttachments(fimMessage);

				verify(fimMessageDataMapper).mapEingangRepresentationsAndAttachments(fimMessage);
			}

			@DisplayName("should return")
			@Test
			void shouldReturn() {
				var result = service.getRepresentationsAndAttachments(fimMessage);

				assertThat(result).isEqualTo(eingang);
			}
		}

		@DisplayName("should throw exception for unexpected message type")
		@Test
		void shouldThrowExceptionForUnexpectedMessageType() {
			var unexpectedMessage = XtaMessageTestFactory.createBuilder()
					.metaData(XtaMessageMetaDataTestFactory.createBuilder()
							.messageTypeCode("unexpected")
							.build())
					.build();

			assertThatThrownBy(() -> service.getRepresentationsAndAttachments(unexpectedMessage))
					.isInstanceOf(TechnicalException.class);
		}
	}

	@DisplayName("process semantik")
	@Nested
	class TestProcessSemantik {
		@Test
		void shouldCallSemantikAdapter() {
			service.processSemantik(FormDataTestFactory.create());

			verify(semantikAdapter).processFormData(any());
		}
	}

}
